from aequitas.plot.commons.style.classes import Shape, Bubble
from aequitas.plot.commons.style import color as Colors
import altair as alt

import math


def get_chart_size_range(length, padding):
    """Calculate the chart size range for a given axis based on the length and padding."""

    return [length * padding, length * (1 - padding)]


def get_color_scale(plot_table, ref_group):
    """
    Calculate the color scale for the bubbles. If there are more groups than the length
    of the categorical color palette, then we use the same color for all the non-reference
    groups. Otherwise, we use an encoding where each group is given its own distinct color.
    The reference group bubble is always given the reference color.
    """
    num_groups = plot_table["attribute_value"].nunique()
    num_colors = len(Colors.CATEGORICAL_COLOR_PALETTE)

    if num_colors >= num_groups - 1:
        color_range = [Colors.REFERENCE] + Colors.CATEGORICAL_COLOR_PALETTE
    else:
        # Non-reference groups are blue. We repeat the color blue per each different group
        color_range = [Colors.REFERENCE] + [
            Colors.BLUE for group in range(1, num_groups)
        ]

    color_domain = list(plot_table["attribute_value"])
    color_domain.insert(0, color_domain.pop(color_domain.index(ref_group)))
    color_scale = alt.Scale(domain=color_domain, range=color_range)
    return color_scale


def get_shape_scale(plot_table, ref_group, accessibility_mode=False):
    """
    Calculate the shape scale for the bubbles centers. The reference group bubble center
    is a cross while the others are circles. If accessibility_mode is on and there are
    more colors in the categorical color palette than different non-reference groups , then we
    also use shapes to distinguish between colors in the scale.
    """

    num_groups = plot_table["attribute_value"].nunique()
    num_colors = len(Colors.CATEGORICAL_COLOR_PALETTE)

    if (num_colors >= num_groups - 1) and accessibility_mode:
        # we find the midpoint in the palette, first half will use the default shape
        # while the second half will use the alternative shape. By the order of the palette,
        # first half should have high constrat between colors, and second half will have
        # some colors that may be similar to the ones in the first half. Therefore, we
        # we change shape to make them more distinguishable in accessibility mode.
        palette_midpoint = math.ceil(num_colors / 2.0)
        shape_range = (
            [Shape.reference]
            + [Shape.default] * palette_midpoint
            + [Shape.alternative] * (num_colors - palette_midpoint)
        )
    else:
        shape_range = [Shape.reference] + [Shape.default] * (num_groups - 1)

    shape_domain = list(plot_table["attribute_value"])
    shape_domain.insert(0, shape_domain.pop(shape_domain.index(ref_group)))
    shape_scale = alt.Scale(domain=shape_domain, range=shape_range)
    return shape_scale


def get_bubble_size_scale(plot_table, metrics, chart_height):
    """Create the scale for the bubble size."""

    max_bubble_radius = (chart_height * Bubble.max_bubble_ratio) / len(metrics)
    bubble_size_domain = [0, plot_table["group_size"].max() * 1.2]
    bubble_size_range = [0, max_bubble_radius ** 2 * math.pi]

    bubble_size_scale = alt.Scale(domain=bubble_size_domain, range=bubble_size_range)
    return bubble_size_scale



import altair as alt
from millify import millify, prettify


def no_axis():
    """Returns an invisible Altair axis."""
    return alt.Axis(domain=False, grid=False, ticks=False, labels=False, title=None)


def format_number(value):
    """Format number to be easily read by humans."""
    if value > 1000000:
        return millify(value, precision=2)
    return prettify(value)


def transform_ratio(value):
    """Transformation that takes ratios and applies a function that preserves equal distances to origin (1)
    for similar relationships, eg. a ratio of 2 (twice the size) is at the same distance of 1 (same size) as
    0.5 (half the size).
    Read: 'how many times larger or smaller than reference disparity'."""
    if value >= 1:
        return value - 1
    else:
        return 1 - 1 / value


def calculate_chart_size_from_elements(header_size, element_size, number_elements):
    """Calculates the total size for a chart based on the desired size for the header (title, axis, etc.), and the
    size and number of elements (lines, columns)."""
    return header_size + (element_size * number_elements)


def to_list(user_input):
    """Wrap a particular `user_input` in a list."""
    return user_input if (isinstance(user_input, list)) else [user_input]


def get_chart_metadata(filename):
    """Generate the metadata for a single Altair chart."""
    return {
        # Vega-Embed (https://github.com/vega/vega-embed) options:
        "embedOptions": {
            # The multiplier for the width and height
            # to export a higher resolution image.
            "scaleFactor": 5,
            # The filename (without extension) for the
            # chart when exported in PNG or SVG format.
            "downloadFileName": filename,
        }
    }



FONT = "Roboto, Arial, sans-serif"

FONT_SIZE_TINY = 8
FONT_SIZE_SMALL = 12
FONT_SIZE_REGULAR = 14
FONT_SIZE_LARGE = 16
FONT_SIZE_EXTRA_LARGE = 20

FONT_WEIGHT_REGULAR = "normal"
FONT_WEIGHT_BOLD = "bold"



import pandas as pd
import altair as alt
import math

from aequitas.plot.commons.style.classes import Legend
from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons.helpers import no_axis
from aequitas.plot.commons import labels as Label


DUMMY_DF = pd.DataFrame({"a": [1, 1], "b": [0, 0]})


def draw_legend(global_scales, selection, chart_width):
    """Draws the interactive group's colors legend for the chart."""

    groups = global_scales["color"].domain
    labels = groups.copy()
    labels[0] = f"{labels[0]} {Label.REF_INDICATOR}"
    legend_df = pd.DataFrame({"attribute_value": groups, "label": labels})

    # Position the legend to the right of the chart

    title_text_x_position = chart_width
    title_text_height = Legend.title_font_size + Legend.row_padding

    radius = math.sqrt(Legend.symbol_size / math.pi)
    entries_circles_x_position = title_text_x_position + radius + Legend.offset
    entries_text_x_position = title_text_x_position + radius * 2 + Legend.offset * 3 - 1

    # Title of the legend.
    title_text = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="left",
            baseline=Legend.title_baseline,
            color=Legend.title_font_color,
            fontSize=Legend.title_font_size,
            font=FONT,
            fontWeight=Legend.title_font_weight,
        )
        .encode(
            x=alt.value(title_text_x_position),
            y=alt.value(Legend.margin),
            text=alt.value(Label.MULTIPLE_GROUPS),
        )
    )

    # Subtitle text that explains how to interact with the legend.
    subtitle_text = (
        alt.Chart(DUMMY_DF)
        .mark_text(
            align="left",
            baseline=Legend.title_baseline,
            color=Legend.title_font_color,
            fontSize=Legend.title_font_size,
            font=FONT,
            fontWeight=Legend.title_font_weight,
        )
        .encode(
            x=alt.value(title_text_x_position),
            y=alt.value(Legend.margin + title_text_height),
            text=alt.value(Label.CLICK_HIGHLIGHT),
        )
    )

    # Conditionally color each legend item
    # If the group is selected, it is colored according to the group color scale, otherwise it is faded
    color_encoding = alt.condition(
        selection,
        alt.Color("attribute_value:N", scale=global_scales["color"], legend=None),
        alt.value(Legend.color_faded),
    )

    # Offset the positioning of the legend items after the subtitle text
    legend_start_y_position = Legend.margin + title_text_height + Legend.font_size + Legend.title_padding

    y_scale = alt.Scale(
        domain=groups,
        range=[
            legend_start_y_position,
            legend_start_y_position
            # number of legend elements x text size
            + (len(groups) * Legend.font_size)
            # number of legend elements x spacing
            + (len(groups) * Legend.offset),
        ],
    )

    # Draw colored symbols for each group
    entries_circles = (
        alt.Chart(legend_df)
        .mark_point(filled=True, opacity=1, size=Legend.symbol_size, cursor=Legend.cursor)
        .encode(
            x=alt.value(entries_circles_x_position),
            y=alt.Y("attribute_value:N", scale=y_scale, axis=no_axis()),
            color=color_encoding,
            shape=alt.Shape(
                "attribute_value:N", scale=global_scales["shape"], legend=None
            ),
        )
        .add_selection(selection)
    )
    trigger_text = alt.selection_multi(empty="all", fields=["attribute_value"])

    # Draw colored label for each group
    entries_text = (
        alt.Chart(legend_df)
        .mark_text(
            align="left",
            baseline="middle",
            font=FONT,
            fontSize=Legend.font_size,
            fontWeight=Legend.font_weight,
            cursor=Legend.cursor,
        )
        .encode(
            x=alt.value(entries_text_x_position),
            y=alt.Y("attribute_value:N", scale=y_scale, axis=no_axis()),
            text=alt.Text("label:N"),
            color=color_encoding,
        )
        .add_selection(trigger_text)
    )

    return entries_circles + entries_text + subtitle_text + title_text



import logging
import uuid
from typing import Any, Dict, List, Tuple, Union
from collections import OrderedDict

import numpy as np
import pandas as pd

logging.getLogger(__name__)

__author__ = (
    "Rayid Ghani, Pedro Saleiro <saleiro@uchicago.edu>, Benedict Kuester, Loren Hinkson, Sérgio Jesus"
)


class Group(object):
    """
    Calculates absolute metrics for a given dataframe with model scores and
    labels, divided by different values on protected attributes.

    Attributes
    ----------
    absolute_metrics : List[str]
        Calculated bias metrics.

    Methods
    -------
    get_crosstabs(df, score_thresholds=None, attr_cols=None, score_col="score", label_col="label_value")
        Calculates metrics for a given dataframe.
    get_multimodel_crosstabs(df, score_thresholds=None, attr_cols=None, score_col="score", label_col="label_value")
        Calculates metrics for a given dataframe with multiple models.
    """

    all_group_metrics = (
        "ppr",
        "pprev",
        "precision",
        "fdr",
        "for",
        "fpr",
        "fnr",
        "tpr",
        "tnr",
        "npv",
        "prev",
    )

    def __init__(self, input_group_metrics=all_group_metrics):
        self.absolute_metrics = input_group_metrics

    @staticmethod
    def gen_df_from_confusion_matrix(
            tp: int,
            tn: int,
            fp: int,
            fn: int,
            k: int,
            model_id: str,
            score_threshold: str,
            attribute_name: str,
            attribute_value: str,
            total_entities: int,
    ):
        columns = [  # Columns in the resulting dataframe
            "model_id",
            "score_threshold",
            "k",
            "attribute_name",
            "attribute_value",
            "tpr",
            "tnr",
            "fomr",
            "fdr",
            "fpr",
            "fnr",
            "npv",
            "precision",
            "pp",
            "pn",
            "ppr",
            "pprev",
            "fp",
            "fn",
            "tn",
            "tp",
            "group_label_pos",
            "group_label_neg",
            "group_size",
            "total_entities",
            "prev",
        ]
        final_dict = {column: [] for column in columns}
        divide = lambda x, y: x / y if y != 0 else np.nan
        # Calculate all metrics from confusion matrix.
        tpr = divide(tp, fn + tp)
        tnr = divide(tn, fp + tn)
        # We can't have variables named 'for', this is changed in return
        fomr = divide(fn, fn + tn)
        fdr = divide(fp, tp + fp)
        fpr = divide(fp, fp + tn)
        fnr = divide(fn, fn + tp)
        npv = divide(tn, fn + tn)
        precision = divide(tp, tp + fp)
        pp = fp + tp
        pn = fn + tn
        ppr = divide(pp, k)
        pprev = divide(pp, pp + pn)
        group_label_pos = fn + tp
        group_label_neg = fp + tn
        group_size = group_label_pos + group_label_neg
        prev = (fn + tp) / group_size
        for key in final_dict.keys():
            # Maybe using locals is not the best, but it is the least
            # messy solution in terms of code.
            final_dict[key].append(locals()[key])
        # Rename column fomr -> for (impossible to have variable named for).
        return pd.DataFrame(final_dict).rename(columns={"fomr": "for"})

    @staticmethod
    def gen_metrics_df(
        df: pd.DataFrame,
        attr_cols: List[str],
        score: str,
        label: str,
        score_threshold: str,
    ) -> pd.DataFrame:
        """
        Generates dataframe with metrics for a given threshold.

        Generates the confusion matrix given the scores in the `score` column
        and the "label_value" for each unique value in each of the `attr_cols`.
        Then, the metrics that are derived from the confusion matrix are
        calculated.

        Parameters
        ----------
        df : pandas.DataFrame
            Classification, label and protected attributes on a given dataset.
            The classification column must already be binary, and in
            column `score`. The label column must be in the column
            "label_value".
        attr_cols : List[str]
            Columns of the dataframe that represent protected attributes.
        score : str
            Column of the dataframe that represents classification result.
            Values in the column must be binary.
        label : str
            Column of the dataframe that represents label of instance.
            Values in the column must be binary.
        score_threshold : str
            Type of threshold used to binarize score.

        Returns
        -------
        pandas.DataFrame
            Metrics for each group in the protected attribute.
        """
        # Method to handle divisions by 0.
        model_id = Group._check_model_id(df, "df")
        # k = Total number of predicted positives in sample.
        # This is for pd.DataFrames only.
        k = df[df[score] == 1].shape[0]
        metrics_dfs = []
        for attribute_name in attr_cols:
            # Create confusion matrix for each group in the form of dictionary.
            # This is for pd.DataFrames only.
            grouped_data = (
                df.groupby(by=[attribute_name, score, label]).size().to_dict(into=OrderedDict)
            )
            # Select the possible (unique) values of the protected attributes (groups).
            # list(OrderedDict.fromkeys()) works like an ordered set.
            attribute_values = list(OrderedDict.fromkeys([key[0] for key in grouped_data]))

            for attribute_value in attribute_values:
                # Get confusion matrix from the dictionary.
                tp = grouped_data.get((attribute_value, 1, 1), 0)
                tn = grouped_data.get((attribute_value, 0, 0), 0)
                fp = grouped_data.get((attribute_value, 1, 0), 0)
                fn = grouped_data.get((attribute_value, 0, 1), 0)
                total_entities = df.shape[0]
                metrics_df = Group.gen_df_from_confusion_matrix(
                    tp, tn, fp, fn, k, model_id, score_threshold, attribute_name,
                    attribute_value, total_entities,
                )
                metrics_dfs.append(metrics_df)
        return pd.concat(metrics_dfs).reset_index().drop(columns="index")

    @staticmethod
    def _check_model_id(df: pd.DataFrame, method_table_name: str = "df") -> int:
        """
        Returns the model id in a dataframe.

        Parameters
        ----------
        df : pandas.DataFrame
            Predictions of a model. Might have or not the column "model_id".
        method_table_name: str
            Reference name of the dataframe for error message.

        Returns
        -------
        int
            The model id. 0 if no model id column exists.
        """
        if "model_id" in df.columns:
            df_models = df.model_id.unique()
            if len(df_models) != 1:
                raise ValueError(
                    "This method requires one and only one model_id in the "
                    "dataframe. "
                    f"Tip: Check that {method_table_name}.model_id.unique() "
                    "only returns an one-element array. "
                )
            else:
                return df_models[0]
        else:
            return 0

    def get_multimodel_crosstabs(
        self,
        df: pd.DataFrame,
        score_thresholds: Dict[str, List[Union[int, float]]] = None,
        attr_cols: List[str] = None,
        score_col: str = "score",
        label_col: str = "label_value",
    ) -> Tuple[pd.DataFrame, List[str]]:
        """
        Performs `get_crosstabs()` and calculates group metrics for results
        from multiple models.

        Parameters
        ----------
        df: pandas.DataFrame
            Results of model classification. Must contain:
            [score_col,  label_col, "model_id"].
        score_thresholds: Dict[str, List[Union[int, float]]]
            The possible thresholds of the models.
            Accepted keys:
                - 'rank_abs' for absolute ranking (top k scores classified as
                   positive).
                - 'rank_pct' for percentage ranking (top X% classified as
                   positive).
                - 'score_val' for value threshold (Scores above Y classified as
                   positive).
            The values for each key represent the different tested thresholds.
            These must be inside a list.
            If not specified, it is assumed the scores are already in binary
            form.
        attr_cols: List[str]
            Columns with protected attributes. If not specified, all
            columns except 'id', 'model_id', 'entity_id', 'score', 'label_value'
            are used. The values in these columns must be categorical.
        score_col : str
            Column of the dataframe that represents classification result.
            Must be numeric / binary.
        label_col : str
            Column of the dataframe that represents label of instance.
            Values in the column must be binary.
        Returns
        -------
        pandas.DataFrame
            Metrics for each group in the protected attribute, for combination
            of model and threshold. List of protected attributes.
        """
        if "model_id" not in df.columns:
            raise ValueError(
                'The method expects a column named "model_id" in the ' "dataframe."
            )

        df_models = df.model_id.unique()
        crosstab_list = []
        model_attr_cols = None
        for model in df_models:
            model_df = df.loc[df["model_id"] == model]
            model_crosstab, model_attr_cols = self.get_crosstabs(
                model_df,
                score_thresholds=score_thresholds,
                attr_cols=attr_cols,
                score_col=score_col,
                label_col=label_col,
            )
            crosstab_list.append(model_crosstab)
        # Note: only returns model_attr_cols from the last iteration, as they
        # all are the same
        return pd.concat(crosstab_list, ignore_index=True), model_attr_cols

    def get_crosstabs(
        self,
        df,
        score_thresholds: Dict[str, List[Union[int, float]]] = None,
        attr_cols: List[str] = None,
        score_col: str = "score",
        label_col: str = "label_value",
    ) -> Tuple[pd.DataFrame, List[str]]:
        """
         Creates univariate groups and calculates group metrics for results
         from a single model.

         Parameters
         ----------
         df : pandas.DataFrame
             Results of model classification. Must contain:
             [score_column,  label_column].
         score_thresholds : Dict[str, List[Union[int, float]]]
             The possible thresholds of the model.
             Accepted keys:
                 - 'rank_abs' for absolute ranking (top k scores classified as
                    positive).
                 - 'rank_pct' for percentage ranking (top X% classified as
                    positive).
                 - 'score_val' for value threshold (Scores above Y classified as
                    positive).
             The values for each key represent the different tested thresholds.
             These must be inside a list.
             If not specified, it is assumed the scores are already in binary
             form.
         attr_cols: List[str]
             Columns with protected attributes. If not specified, all
             columns except 'id', 'model_id', 'entity_id', 'score', 'label_value'
             are used. The values in these columns must be categorical.
         score_col : str
             Column of the dataframe that represents classification result.
             Must be numeric / binary.
         label_col : str
             Column of the dataframe that represents label of instance.
             Values in the column must be binary.
         Returns
         -------
        Tuple[pd.DataFrame, List[str]]
             Metrics for each group in the protected attribute, each specified
             threshold. List of protected attributes.
        """
        if not attr_cols:
            non_attr_cols = ["id", "model_id", "entity_id", score_col, label_col]
            # index of the columns that are protected attributes.
            attr_cols = df.columns[~df.columns.isin(non_attr_cols)].tolist()

        necessary_cols = list(attr_cols) + [score_col, label_col]
        for col in ["id", "model_id", "entity_id"]:
            if col in df.columns:
                necessary_cols.append(col)
        # Copy only the necessary columns
        df = df[necessary_cols].copy()  # To not transform original dataframe.

        # Validation step.
        for col in [score_col, label_col]:
            # Validate if column is in dataframe.
            if col not in df.columns:
                raise KeyError(f'The column "{col}" is not on the dataframe.')
        if len(df[label_col].unique()) > 2:
            raise ValueError("Labels are not binarized.")
        else:
            df[label_col] = df[label_col].astype(int)

        for col in attr_cols:
            # Validate if column is in dataframe.
            if col not in df.columns:
                raise KeyError(f'The attribute column "{col}" is not on the dataframe.')
            # Validate if column has correct datatype.
            if df[col].dtype not in (object, str):
                raise TypeError(
                    f'The attribute column "{col}" has in invalid datatype.'
                )

        # In case no threshold is provided.
        metrics_matrices = []
        if not score_thresholds and len(df[score_col].unique()) > 2:
            raise ValueError("Scores are not binarized. Provide a threshold.")
        elif not score_thresholds:  # No thresholds given and binarized
            df[score_col] = df[score_col].astype(int)
            metrics_matrices.append(
                self.gen_metrics_df(df, attr_cols, score_col, label_col, "binary 0/1")
            )
            score_thresholds = {}
        # In case a threshold is provided.
        binarized_column = str(uuid.uuid4())  # Add a column with binary scores.
        # This is done to maintain the score for future iterations in threshold.
        sorted_df = False  # Flag to sort the dataframe in the first run.
        for key, values in score_thresholds.items():
            for value in values:
                if not sorted_df:
                    # Sort df only once, at the first iteration.
                    df = df.sort_values(score_col, ascending=False).reset_index(
                        drop=True
                    )
                    sorted_df = True
                # Cast scores as ints in the dataframe
                if key == "rank_abs":
                    df[binarized_column] = (df.index < value).astype(int)
                elif key == "rank_pct":
                    df[binarized_column] = (df.index < value * df.shape[0]).astype(int)
                elif key == "score_val":
                    df[binarized_column] = (df[score_col] >= value).astype(int)
                else:
                    raise KeyError("Invalid keys")
                metrics_matrices.append(
                    self.gen_metrics_df(
                        df,
                        attr_cols,
                        binarized_column,
                        label_col,
                        f"{value}_{key[-3:]}",
                    )
                )
        return (pd.concat(metrics_matrices).reset_index().drop(columns="index"),
                attr_cols)

    def list_absolute_metrics(self, df: pd.DataFrame) -> List[Any]:
        """
        View list of all calculated absolute bias metrics in df.

        Parameters
        ----------
        df : pandas.DataFrame
            The result of the `get_crosstabs` methods.

        Returns
        -------
        List[Any]
            Absolute bias metrics.
        """
        return df.columns.intersection(self.absolute_metrics).tolist()



import altair as alt
import pandas as pd

from aequitas.plot.commons.helpers import no_axis, get_chart_metadata
from aequitas.plot.commons.legend import draw_legend
from aequitas.plot.commons.scales import get_chart_size_range
from aequitas.plot.commons.tooltips import get_tooltip_text_group_size
from aequitas.plot.commons.style.classes import (
    Threshold_Band,
    Threshold_Rule,
    Rule,
    Bubble,
    Scatter_Axis,
    Chart_Title,
    Axis,
)

from aequitas.plot.commons.style.text import FONT
from aequitas.plot.commons.style.sizes import XY_Chart
from aequitas.plot.commons import initializers as Initializer
from aequitas.plot.commons import validators as Validator
from aequitas.plot.commons import labels as Label


def __get_position_scales(chart_height, chart_width):
    """Computes the scales for x and y encodings to be used in the xy metrics chart."""

    position_scales = dict()

    # X-Y RANGES (based on chart dimensions)
    x_range = get_chart_size_range(chart_width, XY_Chart.padding_x)
    y_range = get_chart_size_range(chart_height, XY_Chart.padding_y)
    y_range.reverse()

    # METRICS SCALES
    metric_domain = [0, 1]
    position_scales["x"] = alt.Scale(domain=metric_domain, range=x_range)
    position_scales["y"] = alt.Scale(domain=metric_domain, range=y_range)

    return position_scales


def __draw_threshold_bands(
    ref_group_value,
    fairness_threshold,
    main_scale,
    aux_scale,
    accessibility_mode=False,
    drawing_x=False,
):
    """Draws threshold rules and bands for both axis: regions painted red on the chart
    where the metric is above the defined fairness threshold."""

    # DATASOURCE
    threshold_df = pd.DataFrame(
        {
            "start": 0,
            "end": 1,
            "lower_threshold": ref_group_value / fairness_threshold,
            "upper_threshold": min(ref_group_value * fairness_threshold, 1),
        },
        index=[0],
    )

    stroke_color = (
        Threshold_Rule.stroke_accessible
        if accessibility_mode
        else Threshold_Rule.stroke
    )
    fill_color = (
        Threshold_Band.color_accessible if accessibility_mode else Threshold_Band.color
    )

    # BASES
    rule_base = alt.Chart(threshold_df).mark_rule(
        stroke=stroke_color,
        strokeWidth=Threshold_Rule.stroke_width,
        opacity=Threshold_Rule.opacity,
        tooltip=None,
    )

    band_base = alt.Chart(threshold_df).mark_rect(
        fill=fill_color, opacity=Threshold_Band.opacity, tooltip=None
    )

    # PARAMS
    if drawing_x:
        common_params = dict(
            y=alt.Y("start:Q", scale=aux_scale, axis=no_axis()), y2="end:Q"
        )
        upper_param = dict(
            x=alt.X("upper_threshold:Q", scale=main_scale, axis=no_axis())
        )
        lower_param = dict(
            x=alt.X("lower_threshold:Q", scale=main_scale, axis=no_axis())
        )
        lower_end_param = dict(x2="start:Q")
        upper_end_param = dict(x2="end:Q")
    else:
        common_params = dict(
            x=alt.X("start:Q", scale=aux_scale, axis=no_axis()), x2="end:Q"
        )
        upper_param = dict(
            y=alt.Y("upper_threshold:Q", scale=main_scale, axis=no_axis())
        )
        lower_param = dict(
            y=alt.Y("lower_threshold:Q", scale=main_scale, axis=no_axis())
        )
        lower_end_param = dict(y2="start:Q")
        upper_end_param = dict(y2="end:Q")

    # RULES
    lower_threshold_rule = rule_base.encode(
        **common_params,
        **lower_param,
    )
    upper_threshold_rule = rule_base.encode(
        **common_params,
        **upper_param,
    )

    # BANDS
    lower_threshold_band = band_base.encode(
        **common_params, **lower_param, **lower_end_param
    )
    upper_threshold_band = band_base.encode(
        **common_params, **upper_param, **upper_end_param
    )

    return (
        lower_threshold_rule
        + upper_threshold_rule
        + lower_threshold_band
        + upper_threshold_band
    )


def __draw_tick_labels(scales, chart_height, chart_width):
    """Draws the numbers in both axes."""

    axis_values = [0, 0.25, 0.5, 0.75, 1]
    axis_df = pd.DataFrame({"main_axis_values": axis_values, "aux_axis_position": 0})

    x_labels_align_expr = f"datum.main_axis_values === {axis_values[0]} ? 'left' : datum.main_axis_values === {axis_values[-1]} ? 'right' : 'center'"

    x_tick_labels = (
        alt.Chart(axis_df)
        .mark_text(
            baseline="top",
            yOffset=Scatter_Axis.offset,
            tooltip=None,
            align={"expr": x_labels_align_expr},
            fontSize=Scatter_Axis.label_font_size,
            fontWeight=Scatter_Axis.label_font_weight,
            font=FONT,
            color=Scatter_Axis.label_color,
        )
        .encode(
            text=alt.Text("main_axis_values:Q"),
            x=alt.X("main_axis_values:Q", scale=scales["x"], axis=no_axis()),
            y=alt.Y("aux_axis_position:Q", scale=scales["y"], axis=no_axis()),
        )
    )

    y_tick_labels = (
        alt.Chart(axis_df)
        .mark_text(
            baseline="middle",
            xOffset=-Scatter_Axis.offset,
            tooltip=None,
            align="right",
            fontSize=Scatter_Axis.label_font_size,
            fontWeight=Scatter_Axis.label_font_weight,
            font=FONT,
            color=Scatter_Axis.label_color,
        )
        .encode(
            text=alt.Text("main_axis_values:Q"),
            x=alt.X("aux_axis_position:Q", scale=scales["x"], axis=no_axis()),
            y=alt.Y("main_axis_values:Q", scale=scales["y"], axis=no_axis()),
        )
    )

    return x_tick_labels + y_tick_labels


def __draw_axis_rules(x_metric, y_metric, scales):
    """Draws horizontal and vertical rules for the axis."""

    # BASE CHART
    base = alt.Chart(pd.DataFrame({"start": 0, "end": 1}, index=[0]))

    axis_values = [0, 0.25, 0.5, 0.75, 1]
    
    # SHARED PARAMETERS
    grid_expr = f"(datum.value === {axis_values[0]} || datum.value === {axis_values[-1]}) ? 0 : 1"
    shared_axis_params = dict(
        values=axis_values,
        domain=False,
        labels=False,
        ticks=False,
        grid=True,
        gridColor=Axis.grid_color,
        gridCap="round",
        gridDash=[2, 4],
        gridWidth={"expr": grid_expr},
    )

    # AXIS ENCODING
    bottom_axis = alt.Axis(
        **shared_axis_params,
        orient="bottom",
        title=x_metric.upper(),
        titleBaseline="bottom",
        titlePadding=-Scatter_Axis.title_font_size + Scatter_Axis.title_padding,
    )
    left_axis = alt.Axis(
        **shared_axis_params,
        orient="left",
        title=y_metric.upper(),
        titleAlign="left",
        titlePadding=Scatter_Axis.title_font_size / 2 + Scatter_Axis.title_padding,
    )

    # X AXIS
    x_rule = base.mark_rule(
        strokeWidth=1, stroke=Rule.stroke, tooltip=None
    ).encode(
        x=alt.X("start:Q", scale=scales["x"], axis=bottom_axis),
        x2="end:Q",
        y=alt.Y("start:Q", scale=scales["y"], axis=no_axis()),
    )

    # Y AXIS
    y_rule = base.mark_rule(
        strokeWidth=1, stroke=Rule.stroke, tooltip=None
    ).encode(
        x=alt.X("start:Q", scale=scales["x"], axis=no_axis()),
        y=alt.Y("start:Q", scale=scales["y"], axis=left_axis),
        y2="end:Q",
    )

    return x_rule + y_rule


def __draw_bubbles(
    plot_table,
    x_metric,
    y_metric,
    ref_group,
    scales,
    interactive_selection_group,
):
    """Draws the bubbles for all metrics."""

    # FILTER DF
    fields_to_keep_in_metric_table = [
        "group_size",
        "attribute_value",
        "total_entities",
        x_metric,
        y_metric,
    ]
    metric_plot_table = plot_table[fields_to_keep_in_metric_table].copy(deep=True)

    metric_plot_table["tooltip_group_size"] = plot_table.apply(
        lambda row: get_tooltip_text_group_size(
            row["group_size"], row["total_entities"]
        ),
        axis=1,
    )

    # COLOR ENCODING
    bubble_color_encoding = alt.condition(
        interactive_selection_group,
        alt.Color("attribute_value:N", scale=scales["color"], legend=None),
        alt.value(Bubble.color_faded),
    )

    # TOOLTIP ENCODING
    bubble_tooltip_encoding = [
        alt.Tooltip(field="attribute_value", type="nominal", title=Label.SINGLE_GROUP),
        alt.Tooltip(field="tooltip_group_size", type="nominal", title=Label.GROUP_SIZE),
        alt.Tooltip(
            field=x_metric, type="quantitative", format=".2f", title=x_metric.upper()
        ),
        alt.Tooltip(
            field=y_metric, type="quantitative", format=".2f", title=y_metric.upper()
        ),
    ]

    # BUBBLE CENTERS
    bubbles_centers = (
        alt.Chart(metric_plot_table)
        .mark_point(filled=True, size=Bubble.center_size, cursor=Bubble.cursor)
        .encode(
            x=alt.X(f"{x_metric}:Q", scale=scales["x"], axis=no_axis()),
            y=alt.Y(f"{y_metric}:Q", scale=scales["y"], axis=no_axis()),
            tooltip=bubble_tooltip_encoding,
            color=bubble_color_encoding,
            shape=alt.Shape("attribute_value:N", scale=scales["shape"], legend=None),
        )
    )

    # BUBBLE AREAS
    bubbles_areas = (
        alt.Chart(metric_plot_table)
        .mark_circle(opacity=Bubble.opacity, cursor=Bubble.cursor)
        .encode(
            size=alt.Size("group_size:Q", legend=None, scale=scales["bubble_size"]),
            x=alt.X(f"{x_metric}:Q", scale=scales["x"], axis=no_axis()),
            y=alt.Y(f"{y_metric}:Q", scale=scales["y"], axis=no_axis()),
            tooltip=bubble_tooltip_encoding,
            color=bubble_color_encoding,
        )
    )

    return bubbles_centers + bubbles_areas


def plot_xy_metrics_chart(
    disparity_df,
    x_metric,
    y_metric,
    attribute,
    fairness_threshold=1.25,
    chart_height=None,
    chart_width=None,
    accessibility_mode=False,
):
    """Draws XY scatterplot, with the group_size encoded in the bubble size, based on the two metrics provided.
    Additionally draws threshold rules and bands if the fairness_threshold value is not None.

    :param disparity_df: a dataframe generated by the Aequitas Bias class
    :type disparity_df: pandas.core.frame.DataFrame
    :param x_metric: a metric to plot in the x axis
    :type x_metric: str
    :param y_metric: a metric to plot in the y axis
    :type y_metric: str
    :param attribute: an attribute to plot
    :type attribute: str
    :param fairness_threshold: a value for the maximum allowed disparity, defaults to 1.25
    :type fairness_threshold: float, optional
    :param chart_height: a value (in pixels) for the height of the chart
    :type chart_height: int, optional
    :param chart_width: a value (in pixels) for the width of the chart
    :type chart_width: int, optional
    :param accessibility_mode: a switch for the display of more accessible visual elements, defaults to False
    :type accessibility_mode: bool, optional

    :return: the full scatterplot chart
    :rtype: Altair chart object
    """

    (
        plot_table,
        x_metric,
        y_metric,
        ref_group,
        global_scales,
        chart_height,
        chart_width,
        interactive_selection_group,
    ) = Initializer.prepare_xy_chart(
        disparity_df,
        x_metric,
        y_metric,
        attribute,
        fairness_threshold,
        chart_height,
        chart_width,
        XY_Chart,
        accessibility_mode,
    )

    position_scales = __get_position_scales(chart_height, chart_width)

    scales = dict(global_scales, **position_scales)

    # AXIS RULES
    axis_rules = __draw_axis_rules(x_metric, y_metric, scales)

    # TICK LABELS
    tick_labels = __draw_tick_labels(scales, chart_height, chart_width)

    # INITIATE CHART
    chart = axis_rules + tick_labels

    # THRESHOLD AND BANDS
    if fairness_threshold is not None:
        # REF VALUES
        ref_group_index = plot_table.loc[
            plot_table["attribute_value"] == ref_group
        ].index

        x_ref_group_value = plot_table.loc[ref_group_index, x_metric].iloc[0]
        y_ref_group_value = plot_table.loc[ref_group_index, y_metric].iloc[0]

        # Y AXIS
        if y_metric not in Validator.NON_DISPARITY_METRICS_LIST:
            y_thresholds = __draw_threshold_bands(
                ref_group_value=y_ref_group_value,
                fairness_threshold=fairness_threshold,
                main_scale=scales["y"],
                aux_scale=scales["x"],
                accessibility_mode=accessibility_mode,
            )

            chart += y_thresholds

        # X AXIS
        if x_metric not in Validator.NON_DISPARITY_METRICS_LIST:
            x_thresholds = __draw_threshold_bands(
                ref_group_value=x_ref_group_value,
                fairness_threshold=fairness_threshold,
                main_scale=scales["x"],
                aux_scale=scales["y"],
                accessibility_mode=accessibility_mode,
                drawing_x=True,
            )

            chart += x_thresholds

    # LEGEND
    legend = draw_legend(scales, interactive_selection_group, chart_width)

    # BUBBLES
    bubbles = __draw_bubbles(
        plot_table,
        x_metric,
        y_metric,
        ref_group,
        scales,
        interactive_selection_group,
    )

    # FINISH CHART COMPOSITION
    chart += legend + bubbles

    # CONFIGURATION
    styled_chart = (
        chart.configure_view(strokeWidth=0)
        .properties(
            height=chart_height,
            width=chart_width,
            title=f"{y_metric.upper()} by {x_metric.upper()} on {attribute.title()}",
            padding=XY_Chart.full_chart_padding,
            usermeta=get_chart_metadata("scatter_chart"),
        )
        .configure_title(
            font=FONT,
            fontWeight=Chart_Title.font_weight,
            fontSize=Chart_Title.font_size,
            color=Chart_Title.font_color,
            anchor=Chart_Title.anchor,
        )
        .configure_axis(
            titleFont=FONT,
            titleColor=Scatter_Axis.title_color,
            titleFontSize=Scatter_Axis.title_font_size,
            titleFontWeight=Scatter_Axis.title_font_weight,
            titleAngle=Scatter_Axis.title_angle,
        )
        .resolve_scale(y="independent", x="independent", size="independent")
        .resolve_axis(x="shared", y="shared")
    )

    return styled_chart



from aequitas.plot.commons.style import color as Color
from aequitas.plot.commons.style import text as Text


class Bubble:
    opacity = 0.2
    color_faded = Color.FADED
    center_size = 60
    # From bubble size scale: (1 - 2 * chart_padding) / 4
    max_bubble_ratio = 0.2
    cursor = "pointer"


class Title:
    font_size = Text.FONT_SIZE_REGULAR
    font_weight = Text.FONT_WEIGHT_BOLD
    font_color = Color.BLACK
    margin_top = -15


class Chart_Title(Title):
    font_size = Text.FONT_SIZE_EXTRA_LARGE
    font_weight = Text.FONT_WEIGHT_BOLD
    anchor = "start"
    offset = 20


class Subtitle:
    font_size = Text.FONT_SIZE_SMALL
    font_color = Color.GRAY


class Annotation:
    font_size = Text.FONT_SIZE_SMALL
    font_weight = Text.FONT_WEIGHT_REGULAR
    font_color = Color.GRAY
    font_color_reference = Color.REFERENCE
    font_color_threshold = Color.THRESHOLD
    line_spacing = 1.25


class Rule:
    stroke_width = 1.25
    stroke = Color.GRAY


class Reference_Rule(Rule):
    stroke_dash = [5, 5]
    stroke_width = 1
    stroke = Color.REFERENCE


class Threshold_Rule(Rule):
    stroke = Color.THRESHOLD
    stroke_accessible = Color.GRAY
    opacity = 0.8


class Axis(Rule):
    title_color = Color.GRAY
    title_font_size = Text.FONT_SIZE_LARGE
    title_font_weight = Text.FONT_WEIGHT_REGULAR
    label_color = Color.GRAY
    label_font_size = Text.FONT_SIZE_SMALL
    label_font_weight = Text.FONT_WEIGHT_REGULAR
    grid_color = Color.LIGHT_GRAY


class Metric_Axis(Axis):
    offset = 20
    label_font_size = Text.FONT_SIZE_LARGE
    label_padding = -30
    label_padding_concat_chart = -10
    label_angle = 0


class Scatter_Axis(Axis):
    title_padding = 30
    title_font_size = Text.FONT_SIZE_LARGE
    title_font_weight = Text.FONT_WEIGHT_REGULAR
    title_angle = 0
    offset = 6


class Threshold_Band:
    color = Color.THRESHOLD
    color_accessible = Color.GRAY
    opacity = 0.1


class Legend:
    margin = 20
    color_faded = Color.FADED
    font_color = Color.GRAY
    font_size = Text.FONT_SIZE_SMALL
    font_weight = Text.FONT_WEIGHT_REGULAR
    offset = 4
    title_font_color = Color.BLACK
    title_font_size = Text.FONT_SIZE_SMALL
    title_font_weight = Text.FONT_WEIGHT_BOLD
    title_padding = 8
    title_baseline = "top"
    row_padding = 2
    symbol_size = 40
    cursor = "pointer"


class Parity_Result:
    font_size = Text.FONT_SIZE_EXTRA_LARGE
    font_weight = Text.FONT_WEIGHT_BOLD
    color_palette = [Color.REFERENCE] + Color.PARITY_RESULT_PALETTE


class Shape:
    reference = "cross"
    default = "circle"
    alternative = "square"
